#include "lidar_tracker.h"

#include <memory>

namespace lidar_tracker {

std::vector<const ObservationProto*> GetValidObservation(
    const ObservationsProto& observations) {
  std::vector<const ObservationProto*> valid_observations;
  valid_observations.reserve(observations.observations_size());
  for (const auto& obs : observations.observations()) {
    valid_observations.push_back(&obs);
  }
  return valid_observations;
}

std::unique_ptr<Track> LidarTracker::NewTrackFromObservation(
    const ObservationProto& observation) {
  const auto& obs = observation.lidar_observation();
  auto track_ptr = std::make_unique<Track>();
  track_ptr->track_state.id = NewTrackId();
  track_ptr->track_state.timestamp = obs.timestamp();
  track_ptr->track_state.type = obs.type();
  track_ptr->track_state.bounding_box = Box2d(
      {obs.center_x(), obs.center_y()}, obs.yaw(), obs.length(), obs.width());
  track_ptr->track_state.life_state = TrackLifeState::TLS_INIT;
  track_ptr->track_state.estimator.Init(obs.center_x(), obs.center_y(), 0.0,
                                        0.0, obs.timestamp());
  track_ptr->track_state.observation =
      std::make_shared<const ObservationProto>(observation);
  return track_ptr;
}

void LidarTracker::UpdateTrackFromObservation(
    const ObservationProto& observation, Track* track) {
  const double obs_timestamp = observation.timestamp();

  track->track_state.timestamp = obs_timestamp;

  track->track_state.estimator.PredictAndUpdate(observation);
  // TODO(somebody) estimate yaw and shape.

  const Eigen::Vector2d pos = track->track_state.estimator.pos();
  track->track_state.bounding_box =
      Box2d(pos, observation.lidar_observation().yaw(),
            observation.lidar_observation().length(),
            observation.lidar_observation().width());

  track->track_state.observation =
      std::make_shared<const ObservationProto>(observation);

  // TODO(somebody) update lifestate
  // TODO(somebody) update type

  TrackState ckpt = track->track_state;
  track->checkpoints.push_back(std::move(ckpt));
  constexpr double kMaxCheckpointDuration = 1.0;
  while (!track->checkpoints.empty() &&
         track->checkpoints.back().timestamp -
                 track->checkpoints.front().timestamp >
             kMaxCheckpointDuration) {
    track->checkpoints.pop_front();
  }
}

void LidarTracker::Tracking(const ObservationsProto& observations) {
  const auto lidar_observations = GetValidObservation(observations);

  // Association
  const auto match_result = associator_.Associate(tracks_, lidar_observations);

  // Update
  for (int i = 0; i < match_result.size(); ++i) {
    const int track_index = match_result[i];
    const auto* obs = lidar_observations[i];
    if (track_index != -1) {
      // Update track from observation
      UpdateTrackFromObservation(*obs, tracks_[track_index].get());
    } else {
      // Create new track from observation
      tracks_.push_back(NewTrackFromObservation(*obs));
    }
  }

  // Remove tracks
  std::unordered_set<uint32_t> delete_ids;
  for (int i = 0; i < tracks_.size(); ++i) {
    const auto& track = *tracks_[i];
    constexpr double kMaxTrackDuration = 0.3;
    if (observations.has_min_timestamp() &&
        observations.min_timestamp() - track.track_state.timestamp >
            kMaxTrackDuration) {
      delete_ids.insert(track.track_state.id);
    }
  }
  tracks_.erase(
      std::remove_if(tracks_.begin(), tracks_.end(),
                     [&](const auto& track) {
                       return delete_ids.count(track->track_state.id) > 0;
                     }),
      tracks_.end());
}

uint32_t LidarTracker::NewTrackId() { return id_counter_++; }

std::vector<TrackPtr> LidarTracker::GetTracks() const { return tracks_; }

}  // namespace lidar_tracker
